from .template import Template

from .arch import ArchTemplate
from .barrel import BarrelVaultTemplate
from .dome import DomeTemplate
from .wall import WallTemplate

__all__ = [
    "Template",
    "ArchTemplate",
    "BarrelVaultTemplate",
    "DomeTemplate",
    "WallTemplate",
]



from typing import Annotated

from compas.geometry import Frame
from compas.geometry import Line
from compas.geometry import Point
from compas.geometry import Polygon
from compas.geometry import Transformation
from compas.geometry import centroid_points_weighted
from compas.geometry import dot_vectors
from compas.geometry import transform_points
from compas.itertools import pairwise
from compas_model.interactions import Contact


def outer_product(u, v):
    return [[ui * vi for vi in v] for ui in u]


def scale_matrix(M, scale):
    r = len(M)
    c = len(M[0])
    for i in range(r):
        for j in range(c):
            M[i][j] *= scale
    return M


def sum_matrices(A, B):
    r = len(A)
    c = len(A[0])
    M = [[None for j in range(c)] for i in range(r)]
    for i in range(r):
        for j in range(c):
            M[i][j] = A[i][j] + B[i][j]
    return M


class FrictionContact(Contact):
    """Class representing an interaction between two elements through surface-to-surface contact.

    Parameters
    ----------
    forces : list[dict[Literal["c_np", "c_nn", "c_u", "c_v"], float]], optional
        The forces at the corners of the contact.

    Attributes
    ----------
    forces : list[dict[Literal["c_np", "c_nn", "c_u", "c_v"], float]]
        A dictionary of force components per interface point.
        Each dictionary contains the following items: ``{"c_np": ..., "c_nn": ...,  "c_u": ..., "c_v": ...}``.
    points2
    polygon2
    M0
    M1
    M2
    kern
    stressdistribution
    normalforces : list[:class:`compas.geometry.Line`]
        A list of lines representing the normal components of the contact forces at the corners of the interface.
        The length of each line is proportional to the magnitude of the corresponding force.
    compressionforces : list[:class:`compas.geometry.Line`]
        A list of lines representing the compression components of the normal contact forces
        at the corners of the interface.
        The length of each line is proportional to the magnitude of the corresponding force.
    compressiondata
    tensionforces : list[:class:`compas.geometry.Line`]
        A list of lines representing the tension components of the normal contact forces
        at the corners of the interface.
        The length of each line is proportional to the magnitude of the corresponding force.
    tensiondata
    frictionforces : list[:class:`compas.geometry.Line`]
        A list of lines representing the friction or tangential components of the contact forces
        at the corners of the interface.
        The length of each line is proportional to the magnitude of the corresponding force.
    frictiondata
    resultantforce : list[:class:`compas.geometry.Line`]
        A list with a single line representing the resultant of all the contact forces at the corners of the interface.
        The length of the line is proportional to the magnitude of the resultant force.
    resultantdata
    resultantpoint : :class:`compas.geometry.Point`
        The point of application of the resultant force on the interface.

    """

    @property
    def __data__(self) -> dict:
        data = super().__data__
        data["forces"] = self.forces
        return data

    def __init__(self, forces=None, **kwargs):
        super().__init__(**kwargs)

        self._points2 = None
        self._polygon2 = None
        self._forces = forces

        self._compressiondata = None
        self._tensiondata = None
        self._frictiondata = None
        self._resultantdata = None

    # =============================================================================
    # Structural
    # =============================================================================

    @property
    def forces(self) -> list[dict[str, float]]:
        return self._forces

    @forces.setter
    def forces(self, forces: list[dict[str, float]]) -> None:
        self._forces = forces

    @property
    def points2(self) -> list[Point]:
        if not self._points2:
            X = Transformation.from_frame_to_frame(self.frame, Frame.worldXY())
            self._points2 = [Point(*point) for point in transform_points(self.points, X)]
        return self._points2

    @property
    def polygon2(self) -> Polygon:
        if not self._polygon2:
            X = Transformation.from_frame_to_frame(self.frame, Frame.worldXY())
            self._polygon2 = self.polygon.transformed(X)
        return self._polygon2

    @property
    def M0(self) -> float:
        m0 = 0
        for a, b in pairwise(self.points2 + self.points2[:1]):
            d = b - a
            n = [d[1], -d[0], 0]
            m0 += dot_vectors(a, n)
        return 0.5 * m0

    @property
    def M1(self) -> Point:
        m1 = Point(0, 0, 0)
        for a, b in pairwise(self.points2 + self.points2[:1]):
            d = b - a
            n = [d[1], -d[0], 0]
            m0 = dot_vectors(a, n)
            m1 += (a + b) * m0
        return m1 / 6

    @property
    def M2(self) -> Annotated[list[Annotated[list[float], 3]], 3]:
        m2 = outer_product([0, 0, 0], [0, 0, 0])
        for a, b in pairwise(self.points2 + self.points2[:1]):
            d = b - a
            n = [d[1], -d[0], 0]
            m0 = dot_vectors(a, n)
            aa = outer_product(a, a)
            ab = outer_product(a, b)
            ba = outer_product(b, a)
            bb = outer_product(b, b)
            m2 = sum_matrices(
                m2,
                scale_matrix(
                    sum_matrices(sum_matrices(aa, bb), scale_matrix(sum_matrices(ab, ba), 0.5)),
                    m0,
                ),
            )
        return scale_matrix(m2, 1 / 12.0)

    @property
    def kern(self):
        raise NotImplementedError

    @property
    def stressdistribution(self):
        raise NotImplementedError

    @property
    def normalforces(self) -> list[Line]:
        lines = []
        if not self.forces:
            return lines
        frame = self.frame
        w = frame.zaxis
        for point, force in zip(self.points, self.forces):
            force = force["c_np"] - force["c_nn"]
            p1 = point + w * force * 0.5
            p2 = point - w * force * 0.5
            lines.append(Line(p1, p2))
        return lines

    @property
    def compressionforces(self) -> list[Line]:
        lines = []
        if not self.forces:
            return lines
        frame = self.frame
        w = frame.zaxis
        for point, force in zip(self.points, self.forces):
            force = force["c_np"] - force["c_nn"]
            if force > 0:
                p1 = point + w * force * 0.5
                p2 = point - w * force * 0.5
                lines.append(Line(p1, p2))
        return lines

    @property
    def compressiondata(self) -> list[list[float]]:
        if not self._compressiondata:
            self._compressiondata = []
            if self.forces:
                vector = list(self.frame.zaxis)
                for point, force in zip(self.points, self.forces):
                    force = force["c_np"] - force["c_nn"]
                    if force > 0:
                        self._compressiondata.append(list(point) + vector + [0.5 * force])
        return self._compressiondata

    @property
    def tensionforces(self) -> list[Line]:
        lines = []
        if not self.forces:
            return lines
        frame = self.frame
        w = frame.zaxis
        for point, force in zip(self.points, self.forces):
            force = force["c_np"] - force["c_nn"]
            if force < 0:
                p1 = point + w * force * 0.5
                p2 = point - w * force * 0.5
                lines.append(Line(p1, p2))
        return lines

    @property
    def tensiondata(self) -> list[list[float]]:
        if not self._tensiondata:
            self._tensiondata = []
            if self.forces:
                vector = list(self.frame.zaxis)
                for point, force in zip(self.points, self.forces):
                    force = force["c_np"] - force["c_nn"]
                    if force < 0:
                        self._tensiondata.append(list(point) + vector + [0.5 * force])
        return self._tensiondata

    @property
    def frictionforces(self) -> list[Line]:
        lines = []
        if not self.forces:
            return lines
        frame = self.frame
        u, v = frame.xaxis, frame.yaxis
        for point, force in zip(self.points, self.forces):
            ft_uv = (u * force["c_u"] + v * force["c_v"]) * 0.5
            p1 = point + ft_uv
            p2 = point - ft_uv
            lines.append(Line(p1, p2))
        return lines

    @property
    def frictiondata(self) -> list[list[float]]:
        if not self._frictiondata:
            self._frictiondata = []
            if self.forces:
                u, v = list(self.frame.xaxis), list(self.frame.yaxis)
                for point, force in zip(self.points, self.forces):
                    xyz = list(point)
                    self._frictiondata.append(xyz + u + v + [force["c_u"], force["c_v"]])
        return self._frictiondata

    @property
    def resultantpoint(self) -> list[float]:
        if not self.forces:
            return []
        normalcomponents = [f["c_np"] - f["c_nn"] for f in self.forces]
        if sum(normalcomponents):
            return Point(*centroid_points_weighted(self.points, normalcomponents))

    @property
    def resultantforce(self) -> list[Line]:
        if not self.forces:
            return []
        normalcomponents = [f["c_np"] - f["c_nn"] for f in self.forces]
        sum_n = sum(normalcomponents)
        sum_u = sum(f["c_u"] for f in self.forces)
        sum_v = sum(f["c_v"] for f in self.forces)
        position = Point(*centroid_points_weighted(self.points, normalcomponents))
        frame = self.frame
        w, u, v = frame.zaxis, frame.xaxis, frame.yaxis
        forcevector = (w * sum_n + u * sum_u + v * sum_v) * 0.5
        p1 = position + forcevector
        p2 = position - forcevector
        return [Line(p1, p2)]

    @property
    def resultantdata(self) -> list[float]:
        if not self._resultantdata:
            if self.forces:
                normalcomponents = [f["c_np"] - f["c_nn"] for f in self.forces]
                sum_n = sum(normalcomponents)
                sum_u = sum(f["c_u"] for f in self.forces)
                sum_v = sum(f["c_v"] for f in self.forces)
                position = centroid_points_weighted(self.points, normalcomponents)
                u, v, w = self.frame.xaxis, self.frame.yaxis, self.frame.zaxis
                forcevector = u * sum_u + v * sum_v + w * sum_n
                direction = list(forcevector.unitized())
                self._resultantdata = position + direction + [0.5 * forcevector.length]
        return self._resultantdata



from math import cos
from math import pi
from math import sin

from compas.datastructures import Mesh

from .template import Template


def geom_dome(ro, theta, phi):
    x = ro * sin(theta) * cos(phi)
    y = ro * sin(theta) * sin(phi)
    z = ro * cos(theta)
    point = [x, y, z]
    return point


def radius(r_i, r_f, theta_upper, theta_lower, theta):
    r = r_i + (r_f - r_i) / (theta_upper - theta_lower) * theta
    return r


class DomeTemplate(Template):
    """Create voussoirs for a spherical dome geometry with given rise and span."""

    def __init__(
        self,
        meridians=40,
        hoops=20,
        oculus=pi / 30,
        spring=pi / 2,
        r_i=3.9,
        r_f=3.2,
        R_i=4,
        R_f=3.5,
    ):
        super().__init__()
        self.meridians = meridians
        self.hoops = hoops
        self.oculus = oculus
        self.spring = spring

        self.r_i = r_i
        self.r_f = r_f
        self.R_i = R_i
        self.R_f = R_f

    def blocks(self):
        """Compute the blocks.

        Returns
        -------
        list
            A list of blocks defined as simple meshes.

        Notes
        -----
        This method is used by the ``from_geometry`` constructor of the assembly data structure
        to create an assembly "from geometry".

        """
        step_phi = (2 * pi - 0) / (2 * self.meridians)
        phi_delta = (2 * pi - 0) / self.meridians
        theta_delta = (self.spring - self.oculus) / self.hoops

        phi = []
        for i in range(self.meridians + 1):
            phi.append(0 + i * phi_delta)

        theta = []
        for i in range(self.hoops + 1):
            theta.append(self.oculus + i * theta_delta)

        blocks = []

        for j in range(self.meridians):
            for i in range(self.hoops):
                if i % 2 == 0:
                    step = 0
                else:
                    step = step_phi

                vertices = []

                r = radius(self.r_i, self.r_f, self.spring, self.oculus, theta[i])
                R = radius(self.R_i, self.R_f, self.spring, self.oculus, theta[i])

                vertices.append(geom_dome(r, theta[i], phi[j] + step))
                vertices.append(geom_dome(R, theta[i], phi[j] + step))
                vertices.append(geom_dome(R, theta[i], phi[j + 1] + step))
                vertices.append(geom_dome(r, theta[i], phi[j + 1] + step))

                r = radius(self.r_i, self.r_f, self.spring, self.oculus, theta[i + 1])
                R = radius(self.R_i, self.R_f, self.spring, self.oculus, theta[i + 1])

                vertices.append(geom_dome(r, theta[i + 1], phi[j] + step))
                vertices.append(geom_dome(R, theta[i + 1], phi[j] + step))
                vertices.append(geom_dome(R, theta[i + 1], phi[j + 1] + step))
                vertices.append(geom_dome(r, theta[i + 1], phi[j + 1] + step))

                faces = [
                    [0, 4, 5, 1],
                    [1, 5, 6, 2],
                    [0, 1, 2, 3],
                    [0, 3, 7, 4],
                    [5, 4, 7, 6],
                    [6, 7, 3, 2],
                ]

                block = Mesh.from_vertices_and_faces(vertices, faces)
                blocks.append(block)

        return blocks



class Template:
    def __init__(self):
        pass

    def blocks(self):
        raise NotImplementedError

    def interfaces(self):
        raise NotImplementedError

    def to_blocks_and_interfaces(self):
        """Convert the geometry to a list of block meshes,
        and a list of block index pairs representing connections or interfaces.

        Returns
        -------
        tuple
            0. List of meshes representing the block geometries.
            1. List of block index pairs representing connections or interfaces.
        """
        raise NotImplementedError



from math import radians

from compas.datastructures import Mesh
from compas.geometry import Rotation
from compas.geometry import add_vectors
from compas.geometry import angle_vectors
from compas.geometry import subtract_vectors
from compas.geometry import transform_points
from compas.geometry import translate_points

from .template import Template


class BarrelVaultTemplate(Template):
    """Create voussoir geometry for a semi-circular arch with given rise and span.

    Parameters
    ----------
    span : float
        span of the vault
    length : float
        length of the vault perpendicular to the span
    thickness : float
        thickness of the vault
    rise : float
        rise of the vault from 0.0 to middle axis of the vault thickness
    vou_span : int
        number of voussoirs in the span direction
    vou_length : int
        number of voussoirs in the length direction
    zero_is_centerline_or_lowestpoint : bool
        if True, the lowest point of the vault is at the center line of the arch, otherwise the center line of the vault is lowest mesh z-coordinate.

    """

    def __init__(
        self,
        span: float = 6.0,
        length: float = 6.0,
        thickness: float = 0.25,
        rise: float = 0.6,
        vou_span: int = 9,
        vou_length: int = 6,
        zero_is_centerline_or_lowest_point: bool = False,
    ):
        super().__init__()

        self.span = span
        self.length = length
        self.thickness = thickness
        self.rise = rise
        self.vou_span = vou_span
        self.vou_length = vou_length
        self.zero_is_centerline_or_lowest_point = zero_is_centerline_or_lowest_point

    def blocks(self) -> list[Mesh]:
        """Compute the blocks.

        Returns
        -------
        list
            A list of blocks defined as simple meshes.

        """
        span = self.span
        length = self.length
        thickness = self.thickness
        rise = self.rise
        vou_span = self.vou_span
        vou_length = self.vou_length

        radius: float = rise / 2 + span**2 / (8 * rise)
        top: list[float] = [0, 0, rise]
        left: list[float] = [-span / 2, 0, 0]
        center: list[float] = [0.0, 0.0, rise - radius]
        vector: list[float] = subtract_vectors(left, center)
        springing: float = angle_vectors(vector, [-1.0, 0.0, 0.0])
        sector: float = radians(180) - 2 * springing
        angle: float = sector / vou_span

        a: list[float] = [0, -length / 2, rise - (thickness / 2)]
        d: list[float] = add_vectors(top, [0, -length / 2, (thickness / 2)])

        R: Rotation = Rotation.from_axis_and_angle([0, 1.0, 0], 0.5 * sector, center)
        bottom: list[list[float]] = transform_points([a, d], R)
        brick_pts: list[list[list[float]]] = []
        for i in range(vou_span + 1):
            R_angle: Rotation = Rotation.from_axis_and_angle([0, 1.0, 0], -angle * i, center)
            points: list[list[float]] = transform_points(bottom, R_angle)
            brick_pts.append(points)

        depth: float = length / vou_length
        grouped_data: list[list[float]] = [pair[0] + pair[1] for pair in zip(brick_pts, brick_pts[1:])]

        meshes: list[Mesh] = []
        for i in range(vou_length):
            for l, group in enumerate(grouped_data):  # noqa: E741
                is_support: bool = l == 0 or l == (len(grouped_data) - 1)
                if l % 2 == 0:
                    point_l: list[list[float]] = [group[0], group[1], group[2], group[3]]
                    point_list: list[list[float]] = [
                        [group[0][0], group[0][1] + (depth * i), group[0][2]],
                        [group[1][0], group[1][1] + (depth * i), group[1][2]],
                        [group[2][0], group[2][1] + (depth * i), group[2][2]],
                        [group[3][0], group[3][1] + (depth * i), group[3][2]],
                    ]
                    p_t: list[list[float]] = translate_points(point_l, [0, depth * (i + 1), 0])
                    vertices: list[list[float]] = point_list + p_t
                    faces: list[list[int]] = [[0, 1, 3, 2], [0, 4, 5, 1], [4, 6, 7, 5], [6, 2, 3, 7], [1, 5, 7, 3], [2, 6, 4, 0]]
                    mesh: Mesh = Mesh.from_vertices_and_faces(vertices, faces)
                    mesh.attributes["is_support"] = is_support
                    meshes.append(mesh)
                else:
                    point_l: list[list[float]] = [group[0], group[1], group[2], group[3]]
                    points_base: list[list[float]] = translate_points(point_l, [0, depth / 2, 0])
                    points_b_t: list[list[float]] = translate_points(points_base, [0, depth * i, 0])
                    points_t: list[list[float]] = translate_points(points_base, [0, depth * (i + 1), 0])
                    vertices: list[list[float]] = points_b_t + points_t
                    if i != vou_length - 1:
                        faces: list[list[int]] = [[0, 1, 3, 2], [0, 4, 5, 1], [4, 6, 7, 5], [6, 2, 3, 7], [1, 5, 7, 3], [2, 6, 4, 0]]
                        mesh: Mesh = Mesh.from_vertices_and_faces(vertices, faces)
                        mesh.attributes["is_support"] = is_support
                        meshes.append(mesh)

        for l, group in enumerate(grouped_data):  # noqa: E741
            is_support: bool = l == 0 or l == (len(grouped_data) - 1)
            if l % 2 != 0:
                point_l: list[list[float]] = [group[0], group[1], group[2], group[3]]
                p_t: list[list[float]] = translate_points(point_l, [0, depth / 2, 0])
                vertices: list[list[float]] = point_l + p_t
                faces: list[list[int]] = [[0, 1, 3, 2], [0, 4, 5, 1], [4, 6, 7, 5], [6, 2, 3, 7], [1, 5, 7, 3], [2, 6, 4, 0]]
                mesh: Mesh = Mesh.from_vertices_and_faces(vertices, faces)
                mesh.attributes["is_support"] = is_support
                meshes.append(mesh)

                point_f: list[list[float]] = translate_points(point_l, [0, length, 0])
                p_f: list[list[float]] = translate_points(point_f, [0, -depth / 2, 0])
                vertices: list[list[float]] = p_f + point_f
                faces: list[list[int]] = [[0, 1, 3, 2], [0, 4, 5, 1], [4, 6, 7, 5], [6, 2, 3, 7], [1, 5, 7, 3], [2, 6, 4, 0]]
                mesh: Mesh = Mesh.from_vertices_and_faces(vertices, faces)
                mesh.attributes["is_support"] = is_support
                meshes.append(mesh)

        # Find the lowest z-coordinate and move all the block to zero.
        if not self.zero_is_centerline_or_lowest_point:
            min_z: float = min([min(mesh.vertex_coordinates(key)[2] for key in mesh.vertices()) for mesh in meshes])
            for mesh in meshes:
                mesh.translate([0, 0, -min_z])

        return meshes



from typing import Optional

import pythreejs as three
from compas.colors import Color
from compas_notebook.scene import ThreeSceneObject

from compas_dem.models import BlockModel

from .buffers import meshes_to_edgesbuffer
from .buffers import meshes_to_facesbuffer


class ThreeBlockModelObject(ThreeSceneObject):
    """Scene object for drawing mesh."""

    def __init__(
        self,
        show_blocks: Optional[bool] = True,
        show_supports: Optional[bool] = True,
        show_contacts: Optional[bool] = True,
        show_blockfaces: Optional[bool] = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.show_blocks = show_blocks
        self.show_supports = show_supports
        self.show_contacts = show_contacts
        self.show_blockfaces = show_blockfaces

        self.color_edges = Color(0.2, 0.2, 0.2)
        self.color_blocks = Color(0.9, 0.9, 0.9)
        self.color_supports = Color.red().lightened(50)
        self.color_contacts = Color.cyan().lightened(75)

    # @property
    # def settings(self) -> dict:
    #     settings = super().settings
    #     return settings

    @property
    def model(self) -> BlockModel:
        return self.item

    @model.setter
    def model(self, model: BlockModel) -> None:
        self._item = model
        self._transformation = None

    def draw(self):
        """Draw the mesh associated with the scene object.

        Returns
        -------
        list[three.Mesh, three.LineSegments]
            List of pythreejs objects created.

        """
        self._guids = []

        if self.show_blocks:
            facesbuffer, edgesbuffer = self.draw_blocks()
            if facesbuffer:
                self._guids.append(facesbuffer)
            if edgesbuffer:
                self._guids.append(edgesbuffer)

        if self.show_supports:
            facesbuffer, edgesbuffer = self.draw_supports()
            self._guids.append(facesbuffer)
            self._guids.append(edgesbuffer)

        if self.show_contacts:
            facesbuffer, edgesbuffer = self.draw_contacts()
            self._guids.append(facesbuffer)
            self._guids.append(edgesbuffer)

        return self.guids

    def draw_blocks(self) -> tuple[three.Mesh, three.LineSegments]:
        """Draw the blocks of the model."""
        meshes = [block.modelgeometry for block in self.model.blocks()]

        if self.show_blockfaces:
            facesbuffer = meshes_to_facesbuffer(meshes, self.color_blocks)
        else:
            facesbuffer = None

        edgesbuffer = meshes_to_edgesbuffer(meshes, self.color_edges)

        return facesbuffer, edgesbuffer

    def draw_supports(self) -> tuple[three.Mesh, three.LineSegments]:
        """Draw the supports of the model."""
        meshes = [block.modelgeometry for block in self.model.supports()]

        facesbuffer = meshes_to_facesbuffer(meshes, self.color_supports)
        edgesbuffer = meshes_to_edgesbuffer(meshes, self.color_edges)

        return facesbuffer, edgesbuffer

    def draw_contacts(self) -> tuple[three.Mesh, three.LineSegments]:
        """Draw the contacts between the blocks."""
        meshes = [contact.polygon.to_mesh() for contact in self.model.contacts()]

        facesbuffer = meshes_to_facesbuffer(meshes, self.color_contacts)
        edgesbuffer = meshes_to_edgesbuffer(meshes, self.color_edges)

        return facesbuffer, edgesbuffer

    # =============================================================================
    # Individual components
    # =============================================================================

    def draw_blockfaces(self) -> three.Mesh:
        meshes = [block.modelgeometry for block in self.model.blocks()]
        return meshes_to_facesbuffer(meshes, self.color_blocks)

    def draw_blockedges(self) -> three.LineSegments:
        meshes = [block.modelgeometry for block in self.model.blocks()]
        return meshes_to_edgesbuffer(meshes, self.color_edges)

    def draw_supportfaces(self) -> three.Mesh:
        meshes = [block.modelgeometry for block in self.model.supports()]
        return meshes_to_facesbuffer(meshes, self.color_supports)

    def draw_supportedges(self) -> three.LineSegments:
        meshes = [block.modelgeometry for block in self.model.supports()]
        return meshes_to_edgesbuffer(meshes, self.color_edges)

    def draw_contactfaces(self):
        meshes = [contact.polygon.to_mesh() for contact in self.model.contacts()]
        return meshes_to_facesbuffer(meshes, self.color_contacts)

    def draw_contactedges(self):
        meshes = [contact.polygon.to_mesh() for contact in self.model.contacts()]
        return meshes_to_edgesbuffer(meshes, self.color_edges)



import numpy as np
import pythreejs as three
from compas.colors import Color
from compas.datastructures import Mesh
from compas.geometry import Polygon
from compas.geometry import earclip_polygon


def meshes_to_edgesbuffer(
    meshes: list[Mesh],
    color: Color,
) -> three.LineSegments:
    """Convert the combined edges of a collection of meshes to one line segment buffer.

    Parameters
    ----------
    meshes : list[:class:`compas.datastructures.Mesh`]
        The mesh collection.
    color : :class:`compas.colors.Color`
        The color of the edges.

    Returns
    -------
    pythreejs.LineSegments

    """
    positions = []
    colors = []

    for mesh in meshes:
        buffer = mesh_to_edgesbuffer(mesh, color)
        positions += buffer[0]
        colors += buffer[1]

    positions = np.array(positions, dtype=np.float32)
    colors = np.array(colors, dtype=np.float32)

    geometry = three.BufferGeometry(
        attributes={
            "position": three.BufferAttribute(positions, normalized=False),
            "color": three.BufferAttribute(colors, normalized=False, itemSize=3),
        }
    )

    material = three.LineBasicMaterial(vertexColors="VertexColors")

    return three.LineSegments(geometry, material)


def meshes_to_facesbuffer(
    meshes: list[Mesh],
    color: Color,
) -> three.Mesh:
    """Convert the combined faces of a collection of meshes to one mesh buffer.

    Parameters
    ----------
    meshes : list[:class:`compas.datastructures.Mesh`]
        The mesh collection.
    color : :class:`compas.colors.Color`
        The color of the faces.

    Returns
    -------
    pythreejs.Mesh

    """
    positions = []
    colors = []

    for mesh in meshes:
        buffer = mesh_to_facesbuffer(mesh, color)
        positions += buffer[0]
        colors += buffer[1]

    positions = np.array(positions, dtype=np.float32)
    colors = np.array(colors, dtype=np.float32)

    geometry = three.BufferGeometry(
        attributes={
            "position": three.BufferAttribute(positions, normalized=False),
            "color": three.BufferAttribute(colors, normalized=False, itemSize=3),
        }
    )

    material = three.MeshBasicMaterial(
        side="DoubleSide",
        vertexColors="VertexColors",
    )

    return three.Mesh(geometry, material)


def mesh_to_edgesbuffer(mesh: Mesh, color: Color) -> tuple[list[list[float]], list[Color]]:
    positions = []
    colors = []

    for u, v in mesh.edges():
        positions.append(mesh.vertex_coordinates(u))
        positions.append(mesh.vertex_coordinates(v))
        colors.append(color)
        colors.append(color)

    return positions, colors


def mesh_to_facesbuffer(mesh: Mesh, color: Color) -> tuple[list[list[float]], list[Color]]:
    positions = []
    colors = []

    for face in mesh.faces():
        vertices = mesh.face_vertices(face)

        if len(vertices) == 3:
            positions.append(mesh.vertex_coordinates(vertices[0]))
            positions.append(mesh.vertex_coordinates(vertices[1]))
            positions.append(mesh.vertex_coordinates(vertices[2]))
            colors.append(color)
            colors.append(color)
            colors.append(color)

        elif len(vertices) == 4:
            positions.append(mesh.vertex_coordinates(vertices[0]))
            positions.append(mesh.vertex_coordinates(vertices[1]))
            positions.append(mesh.vertex_coordinates(vertices[2]))
            colors.append(color)
            colors.append(color)
            colors.append(color)
            positions.append(mesh.vertex_coordinates(vertices[0]))
            positions.append(mesh.vertex_coordinates(vertices[2]))
            positions.append(mesh.vertex_coordinates(vertices[3]))
            colors.append(color)
            colors.append(color)
            colors.append(color)

        else:
            ears = earclip_polygon(Polygon([mesh.vertex_coordinates(v) for v in vertices]))
            for ear in ears:
                positions.append(mesh.vertex_coordinates(vertices[ear[0]]))
                positions.append(mesh.vertex_coordinates(vertices[ear[1]]))
                positions.append(mesh.vertex_coordinates(vertices[ear[2]]))
                colors.append(color)
                colors.append(color)
                colors.append(color)

    return positions, colors



from typing import Optional

from compas.datastructures import Mesh
from compas.geometry import Box
from compas.geometry import Point
from compas.geometry import Polyhedron
from compas.geometry import Transformation
from compas_model.elements import Element
from compas_model.elements import Feature


# A block could have features like notches,
# but we will work on it when we need it...
# A notch could be a cylinder defined in the frame of a face.
# The frame of a face should be defined in coorination with the global frame of the block.
# during interface detection the features could/should be ignored.
class BlockFeature(Feature):
    pass


class Block(Element):
    """Class representing block elements.

    Parameters
    ----------
    shape : :class:`compas.datastructures.Mesh`
        The base shape of the block.
    features : list[:class:`BlockFeature`], optional
        Additional block features.
    is_support : bool, optional
        Flag indicating that the block is a support.
    frame : :class:`compas.geometry.Frame`, optional
        The coordinate frame of the block.
    name : str, optional
        The name of the element.

    Attributes
    ----------
    shape : :class:`compas.datastructures.Mesh`
        The base shape of the block.
    features : list[:class:`BlockFeature`]
        A list of additional block features.
    is_support : bool
        Flag indicating that the block is a support.

    """

    _geometry: Mesh

    @property
    def __data__(self) -> dict:
        data = super().__data__
        data["geometry"] = self._geometry
        data["is_support"] = self.is_support
        return data

    def __init__(
        self,
        geometry: Mesh,
        features: Optional[list[BlockFeature]] = None,
        transformation: Optional[Transformation] = None,
        name: Optional[str] = None,
        is_support: bool = False,
    ) -> None:
        super().__init__(geometry=geometry, transformation=transformation, features=features, name=name)

        self.is_support = is_support

    # =============================================================================
    # Constructors
    # =============================================================================

    @classmethod
    def from_box(cls, box: Box, **kwargs) -> "Block":
        """Construct a block element from a box.

        Parameters
        ----------
        box : :class:`compas.geometry.Box`
            A box.

        Returns
        -------
        :class:`Block`

        """
        return cls(geometry=Mesh.from_shape(box), **kwargs)

    @classmethod
    def from_polyhedron(cls, polyhedron: Polyhedron, **kwargs) -> "Block":
        """Construct a block element from a polyhedron.

        Parameters
        ----------
        polyhedron : :class:`compas.geometry.Polyhedron`
            A box.

        Returns
        -------
        :class:`Block`

        """
        return cls(geometry=Mesh.from_polyhedron(polyhedron), **kwargs)

    @classmethod
    def from_mesh(cls, mesh: Mesh, **kwargs) -> "Block":
        """Construct a block element from a mesh.

        Parameters
        ----------
        mesh : :class:`compas.datastructures.Mesh`
            A mesh.

        Returns
        -------
        :class:`Block`

        """
        return cls(geometry=mesh.copy(cls=Mesh), **kwargs)

    # =============================================================================
    # Implementations of abstract methods
    # =============================================================================

    def compute_elementgeometry(self, include_features: bool = False) -> Mesh:
        return self._geometry

    def compute_aabb(self, inflate: float = 1.0) -> Box:
        box: Box = self.modelgeometry.aabb()
        if inflate != 1.0:
            box.xsize *= inflate
            box.ysize *= inflate
            box.zsize *= inflate
        self._aabb = box
        return box

    def compute_obb(self, inflate: float = 1.0) -> Box:
        box: Box = self.modelgeometry.obb()
        if inflate != 1.0:
            box.xsize *= inflate
            box.ysize *= inflate
            box.zsize *= inflate
        self._obb = box
        return box

    def compute_point(self) -> Point:
        return Point(*self.modelgeometry.centroid())



from typing import Generator
from typing import Iterator
from typing import Type

from compas.datastructures import Mesh
from compas.geometry import Box
from compas.geometry import Polyhedron
from compas_model.interactions import Contact
from compas_model.models import Model

from compas_dem.elements import Block
from compas_dem.interactions import FrictionContact
from compas_dem.templates import BarrelVaultTemplate
from compas_dem.templates import Template


class BlockModel(Model):
    """Variation of COMPAS Model specifically designed for working with Discrete Element Models in the context of masonry construction."""

    def __init__(self, name=None):
        super().__init__(name)

    def elements(self) -> Iterator[Block]:
        return super().elements()  # type: ignore

    def contacts(self) -> Generator[FrictionContact, None, None]:
        return super().contacts()  # type: ignore

    # =============================================================================
    # Factory methods
    # =============================================================================

    @classmethod
    def from_boxes(cls, boxes: list[Box]) -> "BlockModel":
        """Construct a model from a collection of boxes.

        Parameters
        ----------
        boxes : list[:class:`compas.geometry.Box`]
            A collection of boxes.

        Returns
        -------
        :class:`BlockModel`

        """
        model = cls()
        for box in boxes:
            element = Block.from_box(box)
            model.add_element(element)
        return model

    @classmethod
    def from_polyhedrons(cls, polyhedrons: list[Polyhedron]) -> "BlockModel":
        """Construct a model from a collection of polyhedrons.

        Parameters
        ----------
        polyhedrons : list[:class:`compas.geometry.Polyhedron`]
            A collection of polyhedrons.

        Returns
        -------
        :class:`BlockModel`

        """
        model = cls()
        for polyhedron in polyhedrons:
            element = Block.from_polyhedron(polyhedron)
            model.add_element(element)
        return model

    @classmethod
    def from_polysurfaces(cls, guids):
        """Construct a model from Rhino polysurfaces.

        Parameters
        ----------
        guids : list[str]
            A list of GUIDs identifying the poly-surfaces representing the blocks of the model.

        Returns
        -------
        :class:`BlockModel`

        """
        raise NotImplementedError

    @classmethod
    def from_rhinomeshes(cls, guids):
        """Construct a model from Rhino meshes.

        Parameters
        ----------
        guids : list[str]
            A list of GUIDs identifying the meshes representing the blocks of the model.

        Returns
        -------
        :class:`BlockModel`

        """
        raise NotImplementedError

    @classmethod
    def from_stack(cls):
        raise NotImplementedError

    @classmethod
    def from_wall(cls):
        raise NotImplementedError

    @classmethod
    def from_template(cls, template: Template) -> "BlockModel":
        """Construct a block model from a template.

        Parameters
        ----------
        template : :class:`Template`
            The model template.

        Returns
        -------
        :class:`BlockModel`

        """
        return cls.from_boxes(template.blocks())

    # @classmethod
    # def from_arch(cls):
    #     raise NotImplementedError

    @classmethod
    def from_barrelvault(cls, template: BarrelVaultTemplate):
        """"""
        model = cls()
        for mesh in template.blocks():
            # origin = mesh.face_polygon(5).frame.point
            # frame = Frame(origin, mesh.vertex_point(0) - mesh.vertex_point(2), mesh.vertex_point(4) - mesh.vertex_point(2))
            # xform = Transformation.from_frame_to_frame(frame, Frame.worldXY())
            # mesh_xy: Mesh = mesh.transformed(xform)
            block: Block = Block.from_mesh(mesh)
            # block.is_support = mesh_xy.attributes["is_support"]
            # block.transformation = xform.inverted()
            model.add_element(block)
        return model

    # @classmethod
    # def from_crossvault(cls):
    #     raise NotImplementedError

    # @classmethod
    # def from_fanvault(cls):
    #     raise NotImplementedError

    # @classmethod
    # def from_pavilionvault(cls):
    #     raise NotImplementedError

    @classmethod
    def from_meshpattern(cls):
        raise NotImplementedError

    @classmethod
    def from_nurbssurface(cls):
        raise NotImplementedError

    # =============================================================================
    # Builders
    # =============================================================================

    def add_block_from_mesh(self, mesh: Mesh) -> int:
        block = Block.from_mesh(mesh)
        block.is_support = False
        self.add_element(block)
        return block.graphnode

    def add_support_from_mesh(self, mesh: Mesh) -> int:
        block = Block.from_mesh(mesh)
        block.is_support = True
        self.add_element(block)
        return block.graphnode

    # =============================================================================
    # Blocks & Supports
    # =============================================================================

    def supports(self) -> Generator[Block, None, None]:
        """Iterate over the support blocks of this model.

        Yields
        ------
        :class:`Block`

        """
        for element in self.elements():
            if element.is_support:
                yield element

    def blocks(self) -> Generator[Block, None, None]:
        """Iterate over the regular blocks of this model.

        Yields
        ------
        :class:`Block`

        """
        for element in self.elements():
            if not element.is_support:
                yield element

    # =============================================================================
    # Contacts
    # =============================================================================

    def compute_contacts(
        self,
        tolerance=0.000001,
        minimum_area=0.01,
        contacttype: Type[Contact] = FrictionContact,
    ) -> None:
        return super().compute_contacts(tolerance, minimum_area, contacttype)



from compas_assembly.datastructures import Assembly
from compas_assembly.datastructures import Block
from compas_cra.equilibrium import cra_penalty_solve as _cra_penalty_solve
from compas_cra.equilibrium import rbe_solve as _rbe_solve

from compas_dem.models import BlockModel


def _blockmodel_to_assembly(model: BlockModel) -> Assembly:
    element_block: dict[int, int] = {}

    assembly = Assembly()

    for element in model.elements():
        block: Block = element.modelgeometry.copy(cls=Block)
        x, y, z = element.point
        node = assembly.add_block(block, x=x, y=y, z=z, is_support=element.is_support)
        element_block[element.graphnode] = node

    for edge in model.graph.edges():
        u = element_block[edge[0]]  # type: ignore
        v = element_block[edge[1]]

        contacts = model.graph.edge_attribute(edge, name="contacts")  # type: ignore
        assembly.graph.add_edge(u, v, interfaces=contacts)

    return assembly


def rbe_solve(
    model: BlockModel,
    mu: float = 0.84,
    density: float = 1.0,
    verbose: bool = False,
    timer: bool = False,
):
    assembly = _blockmodel_to_assembly(model)
    _rbe_solve(assembly, mu=mu, density=density, verbose=verbose, timer=timer)


def cra_penalty_solve(
    model: BlockModel,
    mu: float = 0.84,
    density: float = 1.0,
    d_bnd: float = 0.001,
    eps: float = 0.0001,
    verbose: bool = False,
    timer: bool = False,
):
    assembly = _blockmodel_to_assembly(model)
    _cra_penalty_solve(
        assembly,
        mu=mu,
        density=density,
        d_bnd=d_bnd,
        eps=eps,
        verbose=verbose,
        timer=timer,
    )
    for edge in assembly.graph.edges():
        interfaces = assembly.graph.edge_attribute(edge, name="interfaces")
        contacts = model.graph.edge_attribute(edge, name="contacts")
        if interfaces and contacts:
            for interface, contact in zip(interfaces, contacts):
                contact.forces = interface.forces



from .template import Template


class WallTemplate(Template):
    """Create voussoirs for a typical brick wall."""

    def __init__(self):
        super().__init__()

    def blocks(self):
        """Compute the blocks.

        Returns
        -------
        list
            A list of blocks defined as simple meshes.

        Notes
        -----
        This method is used by the ``from_geometry`` constructor of the assembly data structure
        to create an assembly "from geometry".

        """
        pass



from compas.plugins import plugin
from compas.scene import register

from compas_dem.models import BlockModel
from .modelobject import ThreeBlockModelObject


@plugin(category="factories", requires=["pythreejs"])
def register_scene_objects():
    register(BlockModel, ThreeBlockModelObject, context="Notebook")



from math import radians

from compas.datastructures import Mesh
from compas.geometry import Rotation
from compas.geometry import add_vectors
from compas.geometry import angle_vectors
from compas.geometry import subtract_vectors
from compas.geometry import transform_points

from .template import Template


class ArchTemplate(Template):
    """Create voussoir geometry for a semi-circular arch with given rise and span.

    Parameters
    ----------
    rise : float
        The distance between the base of the arch and the highest point of the intrados.
    span : float
        The distance between opposite intrados points at the base.
    thickness : float
        The distance between intrados and extrados.
    depth : float
        The depth of the arch.
    n : int

    """

    def __init__(self, rise, span, thickness, depth, n=None):
        super().__init__()
        self.rise = rise
        self.span = span
        self.thickness = thickness
        self.depth = depth
        self.n = n

    def blocks(self):
        """Compute the blocks.

        Returns
        -------
        list
            A list of blocks defined as simple meshes.

        Notes
        -----
        This method is used by the ``from_geometry`` constructor of the assembly data structure
        to create an assembly "from geometry".

        """
        if self.rise > self.span / 2:
            raise Exception("Not a semicircular arch.")

        radius = self.rise / 2 + self.span**2 / (8 * self.rise)
        # base = [0.0, 0.0, 0.0]
        top = [0.0, 0.0, self.rise]
        left = [-self.span / 2, 0.0, 0.0]
        center = [0.0, 0.0, self.rise - radius]
        vector = subtract_vectors(left, center)
        springing = angle_vectors(vector, [-1.0, 0.0, 0.0])
        sector = radians(180) - 2 * springing
        angle = sector / self.n

        a = top
        b = add_vectors(top, [0, self.depth, 0])
        c = add_vectors(top, [0, self.depth, self.thickness])
        d = add_vectors(top, [0, 0, self.thickness])

        R = Rotation.from_axis_and_angle([0, 1.0, 0], 0.5 * sector, center)
        bottom = transform_points([a, b, c, d], R)

        blocks = []
        for i in range(self.n):
            R = Rotation.from_axis_and_angle([0, 1.0, 0], -angle, center)
            top = transform_points(bottom, R)
            vertices = bottom + top
            faces = [
                [0, 1, 2, 3],
                [7, 6, 5, 4],
                [3, 7, 4, 0],
                [6, 2, 1, 5],
                [7, 3, 2, 6],
                [5, 1, 0, 4],
            ]
            mesh = Mesh.from_vertices_and_faces(vertices, faces)
            blocks.append(mesh)
            bottom = top

        return blocks
